@@ -1,31 +1,34 @@
 import datetime as _datetime
 
+from collections.abc import Mapping
+from typing import IO
 from typing import Tuple
 
 from ._utils import parse_rfc3339
 from .container import Container
 from .items import AoT
-from .items import Comment
-from .items import InlineTable
-from .items import Item as _Item
 from .items import Array
 from .items import Bool
-from .items import Key
+from .items import Comment
 from .items import Date
 from .items import DateTime
 from .items import Float
-from .items import Table
+from .items import InlineTable
 from .items import Integer
+from .items import Item as _Item
+from .items import Key
+from .items import SingleKey
+from .items import String
+from .items import Table
+from .items import Time
 from .items import Trivia
 from .items import Whitespace
-from .items import String
 from .items import item
 from .parser import Parser
-from .toml_document import TOMLDocument as _TOMLDocument
-from .items import Time
+from .toml_document import TOMLDocument
 
 
-def loads(string):  # type: (str) -> _TOMLDocument
+def loads(string: str) -> TOMLDocument:
     """
     Parses a string into a TOMLDocument.
 
@@ -34,48 +37,68 @@ def loads(string):  # type: (str) -> _TOMLDocument
     return parse(string)
 
 
-def dumps(data):  # type: (_TOMLDocument) -> str
+def dumps(data: Mapping, sort_keys: bool = False) -> str:
     """
     Dumps a TOMLDocument into a string.
     """
-    if not isinstance(data, _TOMLDocument) and isinstance(data, dict):
-        data = item(data)
+    if not isinstance(data, Container) and isinstance(data, Mapping):
+        data = item(dict(data), _sort_keys=sort_keys)
 
-    return data.as_string()
+    try:
+        # data should be a `Container` (and therefore implement `as_string`)
+        # for all type safe invocations of this function
+        return data.as_string()  # type: ignore[attr-defined]
+    except AttributeError as ex:
+        msg = f"Expecting Mapping or TOML Container, {type(data)} given"
+        raise TypeError(msg) from ex
 
 
-def parse(string):  # type: (str) -> _TOMLDocument
+def load(fp: IO) -> TOMLDocument:
+    """
+    Load toml document from a file-like object.
+    """
+    return parse(fp.read())
+
+
+def dump(data: Mapping, fp: IO[str], *, sort_keys: bool = False) -> None:
+    """
+    Dump a TOMLDocument into a writable file stream.
+    """
+    fp.write(dumps(data, sort_keys=sort_keys))
+
+
+def parse(string: str) -> TOMLDocument:
     """
     Parses a string into a TOMLDocument.
     """
     return Parser(string).parse()
 
 
-def document():  # type: () -> _TOMLDocument
+def document() -> TOMLDocument:
     """
     Returns a new TOMLDocument instance.
     """
-    return _TOMLDocument()
+    return TOMLDocument()
 
 
 # Items
-def integer(raw):  # type: (str) -> Integer
+def integer(raw: str) -> Integer:
     return item(int(raw))
 
 
-def float_(raw):  # type: (str) -> Float
+def float_(raw: str) -> Float:
     return item(float(raw))
 
 
-def boolean(raw):  # type: (str) -> Bool
+def boolean(raw: str) -> Bool:
     return item(raw == "true")
 
 
-def string(raw):  # type: (str) -> String
+def string(raw: str) -> String:
     return item(raw)
 
 
-def date(raw):  # type: (str) -> Date
+def date(raw: str) -> Date:
     value = parse_rfc3339(raw)
     if not isinstance(value, _datetime.date):
         raise ValueError("date() only accepts date strings.")
@@ -83,7 +106,7 @@ def date(raw):  # type: (str) -> Date
     return item(value)
 
 
-def time(raw):  # type: (str) -> Time
+def time(raw: str) -> Time:
     value = parse_rfc3339(raw)
     if not isinstance(value, _datetime.time):
         raise ValueError("time() only accepts time strings.")
@@ -91,7 +114,7 @@ def time(raw):  # type: (str) -> Time
     return item(value)
 
 
-def datetime(raw):  # type: (str) -> DateTime
+def datetime(raw: str) -> DateTime:
     value = parse_rfc3339(raw)
     if not isinstance(value, _datetime.datetime):
         raise ValueError("datetime() only accepts datetime strings.")
@@ -99,44 +122,44 @@ def datetime(raw):  # type: (str) -> DateTime
     return item(value)
 
 
-def array(raw=None):  # type: (str) -> Array
+def array(raw: str = None) -> Array:
     if raw is None:
         raw = "[]"
 
     return value(raw)
 
 
-def table():  # type: () -> Table
+def table() -> Table:
     return Table(Container(), Trivia(), False)
 
 
-def inline_table():  # type: () -> InlineTable
+def inline_table() -> InlineTable:
     return InlineTable(Container(), Trivia(), new=True)
 
 
-def aot():  # type: () -> AoT
+def aot() -> AoT:
     return AoT([])
 
 
-def key(k):  # type: (str) -> Key
-    return Key(k)
+def key(k: str) -> Key:
+    return SingleKey(k)
 
 
-def value(raw):  # type: (str) -> _Item
+def value(raw: str) -> _Item:
     return Parser(raw)._parse_value()
 
 
-def key_value(src):  # type: (str) -> Tuple[Key, _Item]
+def key_value(src: str) -> Tuple[Key, _Item]:
     return Parser(src)._parse_key_value()
 
 
-def ws(src):  # type: (str) -> Whitespace
+def ws(src: str) -> Whitespace:
     return Whitespace(src, fixed=True)
 
 
-def nl():  # type: () -> Whitespace
+def nl() -> Whitespace:
     return ws("\n")
 
 
-def comment(string):  # type: (str) -> Comment
+def comment(string: str) -> Comment:
     return Comment(Trivia(comment_ws="  ", comment="# " + string))
